#' Creates raw and dissolved segment layers with necessaary
#         upstream/downstream routing attribution
#' @param nhdDF (sf data.frame) valid data frame of NHD flowlines
#' @param POIs  (sf data.frame) Existing POIs
#'
#' @return (sf data.frame) data.frame of segments connecting POIs attributed
#'         with POI_ID for each upstream flowpath
#' @export
make_incremental_segments <- function(nhdDF, POIs, cores = 4) {

  seg_POIs <- dplyr::arrange(POIs, desc(LevelPathI), desc(Hydroseq)) %>%
    dplyr::select(COMID, Hydroseq, LevelPathI) %>%
    dplyr::group_by(LevelPathI) %>%
    # These next two levels arrange POIs hydrologically along the
    #       level path in order to meet the requirements of the code below
    dplyr::mutate(id = row_number(),
           num = if(n() > 1) id[1L] + row_number()-1 else id) %>%
    dplyr::ungroup()

  # Add an empty field for POI_Id population
  nhdDF <- dplyr::mutate(nhdDF, POI_ID = 0)

  POI_ID_assign <- function(i, seg_POIs, nhd){
    ##########################################
    # Populates POI_ID per segment
    #
    # Arguments:
    #   i : (integer) iterator
    #   seg_POIs : (data frame) POI data frame
    #   nhd : (dataframe) flowlines data frame (no geometry)
    #
    # Returns:
    #   nhd_sub : (data frame) raw segments with POI_ID populated
    #
    ##########################################

    # If POI is most upstream POI on levelpath
    if (seg_POIs$num[i] == 1){
      # Assign POI_ID
      nhd_sub <- dplyr::filter(nhd, Hydroseq >= seg_POIs$Hydroseq[i] &
                          LevelPathI == seg_POIs$LevelPathI[i]) %>%
        dplyr::mutate(POI_ID =  seg_POIs$COMID[i])
      # or as you travel downstream on set of POIs below level path
    } else {
      # Assign POI_ID
      nhd_sub <- dplyr::filter(nhd, LevelPathI == seg_POIs$LevelPathI[i] &
                          Hydroseq >= seg_POIs$Hydroseq[i] & Hydroseq < seg_POIs$Hydroseq[i-1]) %>%
        dplyr::mutate(POI_ID = seg_POIs$COMID[i])
    }
    # return assigned flowlines
    return(dplyr::select(nhd_sub, LevelPathI, Hydroseq, COMID, POI_ID) %>%
             dplyr::filter(POI_ID != 0))
  }

  if(!is.null(cores)) {

    clust <- parallel::makeCluster(4)

    POI_list <- parallel::parLapply(clust, c(1:nrow(seg_POIs)), POI_ID_assign, seg_POIs, st_drop_geometry(nhdDF))

    parallel::stopCluster(clust)

  } else {

    POI_list <- lapply(c(1:nrow(seg_POIs)), POI_ID_assign, seg_POIs, st_drop_geometry(nhdDF))

  }

  inc_segs <- data.table::rbindlist(POI_list)

  return(inc_segs)
}
