library(targets)

tar_option_set(packages = c("dplyr", "sf", "hyfabric", "hydroloom", "nhdplusTools"),
               memory = "transient", garbage_collection = TRUE, error = "null", 
               storage = "worker", retrieval = "worker", deployment = "main",
               debug = "wb_inlet_pois_0a67251eed4e40f7")

library(future)
library(future.callr)
plan(multisession)

source("R/config.R")
source("R/user_vars.R")
source("R/utils.R")
source("R/02_POI_creation_functions.R")

prep_path <- ".targets_stores/01_prep/objects/"

list(tar_target(data_paths_file, "cache/data_paths.json", format = "file"),
     tar_target(data_paths, jsonlite::read_json(data_paths_file)),
     tar_target(vpu_codes_file, file.path(prep_path, "vpu_codes"), format = "file"),
     tar_target(vpu_codes, readRDS(vpu_codes_file)),
     tar_target(nav_gpkg, file.path("cache", paste0("reference_", vpu_codes,".gpkg"))),
     tar_target(temp_gpkg, file.path("cache", paste0("nav_", vpu_codes, ".gpkg"))),
     
     ### Base flowline -- gets updated attributes
     tar_target(flowline, read_sf(nav_gpkg, nhd_flowline),
                pattern = map(nav_gpkg), deployment = "worker"),
     
     ### huc12 pois
     tar_target(huc12_poi, create_hu12_pois(read_sf(data_paths$hu12_points_path, "hu_points"),
                                       flowline,
                                       poi_name = "HUC12",
                                       vpu_codes),
                pattern = map(flowline, vpu_codes)),
     
     ### gage POIs
     tar_target(gage_info_gpkg_file, gage_info_gpkg, format = "file"),
     tar_target(gages, read_sf(gage_info_gpkg_file, "Gages")),
     tar_target(gage_pois, create_gage_pois(gages,
                                            flowline, "Gages", vpu_codes, huc12_poi$tmp_POIs, 
                                            min_da_km_gages, combine_meters, reach_meas_thresh),
                pattern = map(flowline, huc12_poi, vpu_codes), deployment = "worker"),
     
     ### Thermo electric power plant POIs
     tar_target(te_points_file, data_paths$TE_points_path, format = "file"),
     tar_target(te_points, read_sf(te_points_file, "2015_TE_Model_Estimates_lat.long_COMIDs")),
     tar_target(te_attributes_file, file.path(data_paths$TE_points_path, "1_model_inputs/te_plants.csv"), format = "file"),
     tar_target(te_attributes, read.csv(te_attributes_file, colClasses = "character")),
     tar_target(te_pois, create_thermo_electric_pois(te_points, te_attributes, 
                                                     gage_pois$tmp_POIs, huc12_poi$HUC12_COMIDs,
                                                     flowline, "TE", vpu_codes),
                pattern = map(gage_pois, huc12_poi, flowline, vpu_codes)),
     
     ### get waterbodies in order
     tar_target(combined_waterbodies_file, file.path(prep_path, "combined_waterbodies"), format = "file"),
     tar_target(combined_waterbodies, readRDS(combined_waterbodies_file)),
     tar_target(waterbodies, make_waterbodies_layer(combined_waterbodies, flowline, geo_crs), 
                pattern = map(flowline), deployment = "worker"),

     ### star fit cross walk and reservoir POIs     
     tar_target(istarf_xwalk_file, data_paths$istarf_xwalk, format = "file"),
     tar_target(istarf_xwalk, read.csv(istarf_xwalk_file)),
     tar_target(resops_pois, create_resops_pois(te_pois$tmp_POIs, waterbodies, istarf_xwalk, 
                                                flowline, "resops"),
                pattern = map(te_pois, waterbodies, flowline), deployment = "worker"),
     ### hilarri is more reservoir POIs
     tar_target(hilarri_file, data_paths$hilarri_sites, format = "file"),
     tar_target(hilarri_data, read.csv(hilarri_file, colClasses = "character")),
     tar_target(hilarri_pois, create_hilarri_pois(resops_pois$tmp_POIs, 
                                                  hilarri_data, 
                                                  resops_pois$wb_pois,
                                                  flowline, "hilarri"),
                pattern = map(resops_pois, flowline), deployment = "worker"),
     
     ### waterbody outlet POIs
     # TODO: work on cleaning up parameters of this function
     # TODO: track event creation explicitly in workflow
     tar_target(wb_outlet_pois, create_wb_outlet_pois(hilarri_pois$tmp_POIs,
                                                      filter(combined_waterbodies, layer == "NHDArea"),
                                                      hilarri_pois$wb_pois,
                                                      resops_pois$wb_resops,
                                                      flowline,
                                                      gage_pois$events,
                                                      nav_gpkg,
                                                      data_paths,
                                                      proj_crs,
                                                      gages,
                                                      "wb_id"),
                pattern = map(hilarri_pois, resops_pois, gage_pois,flowline, nav_gpkg), 
                deployment = "worker"),
     
     ### update flowlines for clarity
     tar_target(updated_flowline_wb_outlet, wb_outlet_pois$updated_flowline, 
                pattern = map(wb_outlet_pois), deployment = "worker"),
     
     ### Addition and Removal events from Brakebill, 2019
     tar_target(ar_data, readxl::read_xlsx(file.path(data_paths$USGS_IT_path, 
                                                       "ARevents_final7.xlsx"))),
     tar_target(ar_event_pois, create_ar_event_pois(wb_outlet_pois$tmp_POIs, ar_data, 
                                                    updated_flowline_wb_outlet, "AR"),
                pattern = map(wb_outlet_pois, updated_flowline_wb_outlet), deployment = "worker"),
     
     ### terminal POIs at network ends
     tar_target(terminal_pois, create_terminal_pois(ar_event_pois$tmp_POIs, 
                                                    updated_flowline_wb_outlet, 
                                                    min_da_km_terminal, "Term"),
                pattern = map(ar_event_pois, updated_flowline_wb_outlet), deployment = "worker"),
     
     ### confluence POIs
     tar_target(confluence_pois, create_confluence_pois(terminal_pois$tmp_POIs, 
                                                        updated_flowline_wb_outlet, "Conf"),
                pattern = map(terminal_pois, updated_flowline_wb_outlet), deployment = "worker"),
     
     ### more updated flowline attributes 
     tar_target(updated_flowline_confluence, confluence_pois$updated_flowline, 
                pattern = map(confluence_pois), deployment = "worker"),
     
     ## waterbody inlet POIs
     tar_target(wb_inlet_pois, create_wb_inlet_pois(confluence_pois$tmp_POIs,
                                                    wb_outlet_pois$WB_VPU_pois,
                                                    wb_outlet_pois$all_events,
                                                    updated_flowline_confluence, gages,
                                                    data_paths, proj_crs, "WBIn"),
                pattern = map(confluence_pois, wb_outlet_pois, updated_flowline_confluence), deployment = "worker")
     )